@@ -1,18 +1,7 @@
 use std::convert::TryFrom;
 
-use crate::{
-    immutable::{object::decode, parse::SPACE},
-    tree,
-};
-use bstr::{BStr, ByteSlice};
-use nom::{
-    bytes::complete::{tag, take, take_while1, take_while_m_n},
-    character::is_digit,
-    combinator::all_consuming,
-    multi::many0,
-    sequence::terminated,
-    IResult,
-};
+use crate::{immutable::object, tree};
+use bstr::BStr;
 
 /// A directory snapshot containing files (blobs), directories (trees) and submodules (commits).
 #[derive(PartialEq, Eq, Debug, Hash, Ord, PartialOrd, Clone)]
@@ -54,8 +43,8 @@ pub struct Entry<'a> {
 
 impl<'a> Tree<'a> {
     /// Deserialize a Tree from `data`.
-    pub fn from_bytes(data: &'a [u8]) -> Result<Tree<'a>, decode::Error> {
-        parse(data).map(|(_, t)| t).map_err(decode::Error::from)
+    pub fn from_bytes(data: &'a [u8]) -> Result<Tree<'a>, object::decode::Error> {
+        decode::tree(data).map(|(_, t)| t).map_err(object::decode::Error::from)
     }
 
     /// Create an instance of the empty tree.
@@ -68,7 +57,7 @@ impl<'a> Tree<'a> {
 
 impl<'a> TreeIter<'a> {
     /// Consume self and return all parsed entries.
-    pub fn entries(self) -> Result<Vec<Entry<'a>>, decode::Error> {
+    pub fn entries(self) -> Result<Vec<Entry<'a>>, object::decode::Error> {
         self.collect()
     }
 }
@@ -80,13 +69,13 @@ impl<'a> Default for TreeIter<'a> {
 }
 
 impl<'a> Iterator for TreeIter<'a> {
-    type Item = Result<Entry<'a>, decode::Error>;
+    type Item = Result<Entry<'a>, object::decode::Error>;
 
     fn next(&mut self) -> Option<Self::Item> {
         if self.data.is_empty() {
             return None;
         }
-        match parse_entry(self.data) {
+        match decode::entry(self.data) {
             Ok((data_left, entry)) => {
                 self.data = data_left;
                 Some(Ok(entry))
@@ -99,10 +88,10 @@ impl<'a> Iterator for TreeIter<'a> {
     }
 }
 
-impl TryFrom<&[u8]> for tree::EntryMode {
-    type Error = decode::Error;
+impl<'a> TryFrom<&'a [u8]> for tree::EntryMode {
+    type Error = &'a [u8];
 
-    fn try_from(mode: &[u8]) -> Result<Self, Self::Error> {
+    fn try_from(mode: &'a [u8]) -> Result<Self, Self::Error> {
         Ok(match mode {
             b"40000" => tree::EntryMode::Tree,
             b"100644" => tree::EntryMode::Blob,
@@ -111,29 +100,50 @@ impl TryFrom<&[u8]> for tree::EntryMode {
             b"100755" => tree::EntryMode::BlobExecutable,
             b"120000" => tree::EntryMode::Link,
             b"160000" => tree::EntryMode::Commit,
-            _ => return Err(decode::Error::NomDetail(mode.into(), "unknown tree mode")),
+            _ => return Err(mode),
         })
     }
 }
 
-const NULL: &[u8] = b"\0";
-fn parse_entry(i: &[u8]) -> IResult<&[u8], Entry<'_>, decode::Error> {
-    let (i, mode) = terminated(take_while_m_n(5, 6, is_digit), tag(SPACE))(i)?;
-    let mode = tree::EntryMode::try_from(mode).map_err(nom::Err::Error)?;
-    let (i, filename) = terminated(take_while1(|b| b != NULL[0]), tag(NULL))(i)?;
-    let (i, oid) = take(20u8)(i)?;
+mod decode {
+    use std::convert::TryFrom;
 
-    Ok((
-        i,
-        Entry {
-            mode,
-            filename: filename.as_bstr(),
-            oid: git_hash::oid::try_from(oid).expect("we counted exactly 20 bytes"),
-        },
-    ))
-}
+    use crate::{
+        immutable::{parse::SPACE, tree::Entry, Tree},
+        tree,
+    };
+    use bstr::ByteSlice;
+    use nom::{
+        bytes::complete::{tag, take, take_while1, take_while_m_n},
+        character::is_digit,
+        combinator::all_consuming,
+        error::ParseError,
+        multi::many0,
+        sequence::terminated,
+        IResult,
+    };
 
-fn parse(i: &[u8]) -> IResult<&[u8], Tree<'_>, decode::Error> {
-    let (i, entries) = all_consuming(many0(parse_entry))(i)?;
-    Ok((i, Tree { entries }))
+    const NULL: &[u8] = b"\0";
+
+    pub fn entry<'a, E: ParseError<&'a [u8]>>(i: &'a [u8]) -> IResult<&[u8], Entry<'_>, E> {
+        let (i, mode) = terminated(take_while_m_n(5, 6, is_digit), tag(SPACE))(i)?;
+        let mode = tree::EntryMode::try_from(mode)
+            .map_err(|invalid| nom::Err::Error(E::from_error_kind(invalid, nom::error::ErrorKind::MapRes)))?;
+        let (i, filename) = terminated(take_while1(|b| b != NULL[0]), tag(NULL))(i)?;
+        let (i, oid) = take(20u8)(i)?;
+
+        Ok((
+            i,
+            Entry {
+                mode,
+                filename: filename.as_bstr(),
+                oid: git_hash::oid::try_from(oid).expect("we counted exactly 20 bytes"),
+            },
+        ))
+    }
+
+    pub fn tree<'a, E: ParseError<&'a [u8]>>(i: &'a [u8]) -> IResult<&'a [u8], Tree<'a>, E> {
+        let (i, entries) = all_consuming(many0(entry))(i)?;
+        Ok((i, Tree { entries }))
+    }
 }
